# Motion class.


class Motion:
    def __init__(self, motion_type, motion_direction, motion_origin_pos, motion_viz_orient, motion_dir_color, motion_origin_color, visible):
        self.motion_type = motion_type
        self.motion_direction = motion_direction
        self.motion_origin_pos = motion_origin_pos
        self.motion_viz_orient = motion_viz_orient
        self.motion_dir_color = motion_dir_color
        self.motion_origin_color = motion_origin_color
        self.visible = visible
    
    def get_properties(self, binary_filename):
        """ 
        :return: A dict containing object properties. They are written into json and interpreted by javascript.
        """
        json_dict = {
            'type': 'motion',
            'motion_type': self.motion_type,
            'motion_direction': self.motion_direction.tolist(),
            'motion_origin_pos': self.motion_origin_pos.tolist(),
            'motion_viz_orient': self.motion_viz_orient,
            'motion_dir_color': self.motion_dir_color.tolist(),
            'motion_origin_color': self.motion_origin_color.tolist(),
            'visible': self.visible,
        }
        return json_dict

    def write_binary(self, path):
        """Write lines to binary file."""
        return



"""Labels class e.g. to visualize the labels of instances."""
import numpy as np


class Labels:
    """Set of labels."""

    def __init__(self, labels, positions, colors, visible):
        self.labels = labels
        self.positions = positions
        self.colors = colors
        self.visible = visible

    def get_properties(self, binary_filename):
        """ Get arrow properties, they are written into json and interpreted by javascript.
        :return: A dict conteining object properties.
        """

        positions = np.array(self.positions)
        colors = np.array(self.colors)

        json_dict = {
            'type': 'labels',
            'labels': self.labels,
            'positions': positions.tolist(),
            'colors': colors.tolist(),
            'visible': self.visible,
        }
        return json_dict

    def write_binary(self, path):
        """Write lines to binary file."""
        return

    def write_blender(self, path):
        print(type(self).__name__+'.write_blender() not yet implemented.' )
        return


# Arrow class.


class Arrow:
    def __init__(self, start, end, color, alpha, stroke_width, head_width, visible):
        self.start = start
        self.end = end
        self.color = color
        self.alpha = alpha
        self.stroke_width = stroke_width
        self.head_width = head_width
        self.visible = visible

    def get_properties(self, binary_filename):
        """ Get arrow properties, they are written into json and interpreted by javascript.
        :return: A dict conteining object properties.
        """
        json_dict = {
            'type': 'arrow',
            'start': self.start.tolist(),
            'end': self.end.tolist(),
            'color': self.color.tolist(),
            'alpha': float(self.alpha),
            'stroke_width': float(self.stroke_width),
            'head_width': float(self.head_width),
            'visible': self.visible,
        }
        return json_dict

    def write_binary(self, path):
        """Write lines to binary file."""
        return

    def write_blender(self, path):
        print(type(self).__name__ + '.write_blender() not yet implemented.' )
        return


"""Polygon mesh."""
from shutil import copyfile
import os

class Mesh:
    """Polygon mesh defined by .obj file and some solid color."""

    def __init__(self, filename, translation=None, rotation=None, scale=None, color=None, visible=True):
        self.filename_source = filename
        mesh_file = os.path.split(filename)[-1]
        mesh_file_name = mesh_file.split('.')[0]
        self.mesh_file_extension = mesh_file.split('.')[1]
        mesh_file_size = os.path.getsize(filename)
        self.filename_destination = mesh_file_name + '_' + str(mesh_file_size) + '.' + self.mesh_file_extension
        self.translation = translation.tolist()
        self.rotation = rotation.tolist()
        self.scale = scale.tolist()
        self.color = color
        self.visible = visible

    def get_properties(self, filename):
        """
        :return: A dict containing object properties. They are written into json and interpreted by javascript.
        """
        json_dict = {
            'type': 'mesh',
            'filename': self.filename_destination,
            'translation': self.translation,
            'rotation': self.rotation,
            'scale': self.scale,
            'visible': self.visible,
            }
        if self.color is not None:
            json_dict['color'] = self.color.tolist()
        return json_dict

    def write_binary(self, path):
        destination_dir = os.path.dirname(path)
        destination_path = os.path.join(destination_dir, self.filename_destination)
        if not os.path.exists(self.filename_destination):
            copyfile(self.filename_source, destination_path)

    def write_blender(self, path):
        return


"""The visualizer class is used to show 3d scenes."""

from .points import Points
from .labels import Labels
from .lines import Lines
from .mesh import Mesh
from .camera import Camera
from .cuboid import Cuboid
from .polyline import Polyline
from .arrow import Arrow
from .circles_2d import Circles2D
from .motion import Motion

import os
import sys
import shutil
import json
import numpy as np

def euler_to_quaternion(x: float, y: float, z: float):
    cr = np.cos(x * 0.5)
    sr = np.sin(x * 0.5)
    cp = np.cos(y * 0.5)
    sp = np.sin(y * 0.5)
    cy = np.cos(z * 0.5)
    sy = np.sin(z * 0.5)
    q = np.zeros([4])
    q[0] = sr * cp * cy - cr * sp * sy  # x
    q[1] = cr * sp * cy + sr * cp * sy  # y
    q[2] = cr * cp * sy - sr * sp * cy  # z
    q[3] = cr * cp * cy + sr * sp * sy  # w
    return q

class Visualizer:
    def __init__(self,
                 position: np.array = np.array([3.0, 3.0, 3.0]),
                 look_at: np.array = np.array([0.0, 0.0, 0.0]),
                 up: np.array = np.array([0.0, 0.0, 1.0]),
                 focal_length: float = 28.0, animation=False):

        self.camera = Camera(
            position=np.array(position),
            look_at=np.array(look_at),
            up=np.array(up),
            focal_length=focal_length,
            animation=animation
        )
        self.elements = {"Camera_0": self.camera}

    def __parse_name(self,
                     name: str) -> str:
        """Makes sure the name does not contain invalid character combinations.

        :param name:
        :return:
        """
        return name.replace(':', ';')

    def save(self,
            path: str,
            port: int=6008,
            blender_args: dict=None,
            verbose=True):
        """Creates the visualization and displays the link to it.

        :param path: The path to save the visualization files.
        :param port: The port to show the visualization.
        :param verbose: Whether to print the web-server message or not.
        """

        # Delete destination directory if it exists already
        directory_destination = os.path.abspath(path)
        if os.path.isdir(directory_destination):
            shutil.rmtree(directory_destination)

        # Copy website directory
        directory_source = os.path.realpath(os.path.join(os.path.dirname(__file__), "src"))
        shutil.copytree(directory_source, directory_destination)

        # Assemble binary data files
        nodes_dict = {}
        for name, e in self.elements.items():
            binary_file_path = os.path.join(directory_destination, name + ".bin")
            nodes_dict[name] = e.get_properties(name + ".bin")
            e.write_binary(binary_file_path)
            if blender_args:
                blender_file_oath = os.path.join(directory_destination, name + ".ply")
                e.write_blender(blender_file_oath)

        # Write json file containing all scene elements
        json_file = os.path.join(directory_destination, "nodes.json")
        with open(json_file, "w") as outfile:
            json.dump(nodes_dict, outfile)

        if not verbose:
            return

        # Display link
        http_server_string = "python -m SimpleHTTPServer " + str(port)
        if sys.version[0] == "3":
            http_server_string = "python -m http.server " + str(port)
        print("")
        print(
            "************************************************************************"
        )
        print("1) Start local server:")
        print("    cd " + directory_destination + "; " + http_server_string)
        print("2) Open in browser:")
        print("    http://localhost:" + str(port))
        print(
            "************************************************************************"
        )

        if blender_args:
            self.show_in_blender(path, blender_args, verbose)

    def show_in_blender(self,
                        path: str,
                        blender_args: dict,
                        verbose: bool=True):

        directory_destination = os.path.abspath(path)
        blender_script_path = os.path.join(directory_destination, "blender_script.py")
        with open(blender_script_path, "w") as outfile:
            outfile.write(
"import bpy\nimport os\n\
import sys\n\
sys.path.append(os.getcwd())\n\
import blender_tools\n\
blender_tools.main()")

        cmd = "cd " + directory_destination + "; " + blender_args['executable_path'] + " --background --python blender_script.py"
        if blender_args['output_prefix']:
            cmd = cmd + " -- " + blender_args['output_prefix']
        os.system(cmd)

        if verbose:
            print("")
            print("************************************************************************")
            print("Blender instructions")
            print(cmd)
            print("************************************************************************")
        
    def add_points(
        self,
        name: str,
        positions: np.array,
        colors: np.array=None,
        normals: np.array=None,
        point_size: int=25,
        resolution: int=5,
        visible: bool=True,
        alpha: float=1.0,
    ):
        """Add points to the visualizer.

        :param name: The name of the points displayed in the visualizer. Use ';' in the name to create sub-layers.
        :param positions: The point positions.
        :param normals: The point normals.
        :param colors: The point colors.
        :param point_size: The point size.
        :param resolution: The resolution of the blender sphere.
        :param visible: Bool if points are visible.
        :param alpha: Alpha value of colors.
        """

        assert positions.shape[1] == 3
        assert colors is None or positions.shape == colors.shape
        assert normals is None or positions.shape == normals.shape

        shading_type = 1  # Phong shading
        if colors is None:
            colors = np.ones(positions.shape, dtype=np.uint8) * 50  # gray
        if normals is None:
            normals = np.ones(positions.shape, dtype=np.float32)
            shading_type = 0  # Uniform shading when no normals are available

        positions = positions.astype(np.float32)
        colors = colors.astype(np.uint8)
        normals = normals.astype(np.float32)

        alpha = min(max(alpha, 0.0), 1.0)  # cap alpha to [0..1]

        self.elements[self.__parse_name(name)] = Points(
            positions, colors, normals, point_size, resolution, visible, alpha, shading_type
        )

    def add_labels(self,
                   name: str,
                   labels: list,
                   positions: np.array,
                   colors: np.array,
                   visible: bool=True):
        """Add labels to the visualizer.
        
        :param name: The name of the labels.
        :param labels: The text value of the labels.
        :param positions: The 3D positions of the labels.
        :param colors: The text color of the individual labels.
        :param visible: Bool if label is visible.
        """
        self.elements[self.__parse_name(name)] = Labels(labels, positions, colors, visible)

    def add_circles_2d(self,
                       name: str,
                       labels: list,
                       positions: np.array,
                       border_colors: np.array,
                       fill_colors: np.array,
                       visible: bool=True):
        """Add node to the visualizer.
        
        :param name: The name of the node.
        :param labels: The text value of the node.
        :param positions: The 3D positions of the node.
        :param border_colors: The text color of the individual node.
        :param fill_colors: The text color of the individual node.
        :param visible: Bool if lines are visible.
        """
        self.elements[self.__parse_name(name)] = Circles2D(labels, positions, border_colors, fill_colors, visible)

    def add_lines(self,
                  name: str,
                  lines_start: np.array,
                  lines_end: np.array,
                  colors: np.array=None,
                  visible: bool=True):
        """Add lines to the visualizer.

        :param name: The name of the lines displayed in the visualizer.
        :param lines_start: The start positions of the lines.
        :param lines_end: The end positions of the lines.
        :param colors: The line colors.
        :param visible: Bool if lines are visible.
        """

        assert lines_start.shape[1] == 3
        assert lines_start.shape == lines_end.shape
        assert colors is None or lines_start.shape == colors.shape

        if colors is None:
            colors = np.ones(lines_start.shape, dtype=np.uint8) * 50  # gray

        colors = colors.astype(np.uint8)
        lines_start = lines_start.astype(np.float32)
        lines_end = lines_end.astype(np.float32)
        self.elements[self.__parse_name(name)] = Lines(lines_start, lines_end, colors, colors, visible)

    def add_bounding_box(self,
                         name: str,
                         position: np.array,
                         size: np.array,
                         rotation: np.array=np.array([0.0, 0.0, 0.0, 1.0]),
                         color: np.array=np.array([255, 0, 0]),
                         alpha: float=1.0,
                         edge_width: float=0.01,
                         visible: bool=True):
        """Add oriented 3D bounding box."""
        rotation /= np.linalg.norm(rotation)  # normalize the orientation
        self.elements[self.__parse_name(name)] = Cuboid(position, size, rotation, color, alpha, edge_width, visible)

    def add_mesh(self,
                 name: str,
                 path: str,
                 translation: np.array=np.array([0.0, 0.0, 0.0]),
                 rotation: np.array=np.array([0.0, 0.0, 0.0, 1.0]),  # [x, y, z, w] - rotate w degrees rad around the axis xyz
                 scale: np.array=np.array([1, 1, 1]),
                 color: np.array=None,
                 visible: bool=True):
        """Adds a polygon mesh to the scene as specified in the path.
         
          The path is currently limited to .obj files and the color is the uniform color of the objetc.
        """
        rotation /= np.linalg.norm(rotation)  # normalize the orientation
        self.elements[self.__parse_name(name)] = Mesh(path, translation=translation, rotation=rotation, scale=scale, color=color, visible=visible)

    def add_polyline(self,
                     name: str,
                     positions: np.array,
                     color: np.array=np.array([255, 0, 0]),
                     alpha: float=1.0,
                     edge_width: float=0.01,
                     visible: bool=True):
        """Add polyline.

        :param name: The bounding box name. (string)
        :param positions: The N 3D positions along the polyline. (float32, Nx3)
        :param color: The color. (int32, 3x1)
        :param alpha: The transparency. (float32)
        :param edge_width: The width of the edges. (float32)
        :param visible: Bool, whether visible or not.
        """

        self.elements[self.__parse_name(name)] = Polyline(positions, color, alpha, edge_width, visible)

    def add_superquadric(self,
        name: str,
        scalings: np.array=np.array([1.0, 1.0, 1.0]),
        exponents: np.array=np.array([2.0, 2.0, 2.0]),
        translation: np.array=np.array([0.0, 0.0, 0.0]),
        rotation: np.array=np.array([1.0, 0.0, 0.0, 0.0]),
        color: np.array=np.array([255, 255, 255]),
        resolution: int=30,
        visible: bool=True):
        """Adds a superqiadroc mesh to the scene."""

        def create_superquadric_mesh(A, B, C, r, s, t, N):
            def f(o, m):
                return np.sign(np.sin(o)) * np.abs(np.sin(o))**m
            def g(o, m):
                return np.sign(np.cos(o)) * np.abs(np.cos(o))**m
            u = np.linspace(-np.pi, np.pi, N, endpoint=True)
            v = np.linspace(-np.pi/2.0, np.pi/2.0, N, endpoint=True)
            u = np.tile(u, N)
            v = np.repeat(v, N)
            triangles = []

            x = A * g(v, 2.0 / r) * g(u, 2.0 / r)
            y = B * g(v, 2.0 / s) * f(u, 2.0 / s)
            z = C * f(v, 2.0 / t)
            vertices =  np.concatenate([np.expand_dims(x, 1),
                                        np.expand_dims(y, 1),
                                        np.expand_dims(z, 1)], axis=1)
            triangles = []
            for i in range(N-1):
                for j in range(N-1):
                    triangles.append([i*N+j, (i+1)*N+j+1, (i+1)*N+j])
                    triangles.append([i*N+j, i*N+(j+1), (i+1)*N+(j+1)])
            return vertices, triangles

        vertices, triangles = create_superquadric_mesh(scalings[0], scalings[1], scalings[2],
                                                       exponents[0], exponents[1], exponents[2],
                                                       resolution)
        import open3d as o3d
        mesh_sq = o3d.geometry.TriangleMesh()
        mesh_sq.vertices = o3d.utility.Vector3dVector(vertices)
        mesh_sq.triangles = o3d.utility.Vector3iVector(triangles)
        o3d.io.write_triangle_mesh(f"{name}.obj", mesh_sq, write_ascii=True, compressed=False, write_vertex_normals=False, write_vertex_colors=False, write_triangle_uvs=False, print_progress=False)

        rotation /= np.linalg.norm(rotation)  # normalize the orientation
        scale = np.array([1.0, 1.0, 1.0])
        self.elements[self.__parse_name(name)] = Mesh(f"{name}.obj", translation=translation, rotation=rotation, scale=scale, color=color, visible=visible)

    def add_superquadric_rot_mat(self,
        name: str,
        scalings: np.array=np.array([1.0, 1.0, 1.0]),
        exponents: np.array=np.array([2.0, 2.0, 2.0]),
        translation: np.array=np.array([0.0, 0.0, 0.0]),
        rotation: np.array=np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0],[0.0, 0.0,1.0]]),
        color: np.array=np.array([255, 255, 255]),
        resolution: int=30,
        visible: bool=True):
        """Adds a superqiadroc mesh to the scene."""

        def create_superquadric_mesh(A, B, C, r, s, t, N):
            def f(o, m):
                return np.sign(np.sin(o)) * np.abs(np.sin(o))**m
            def g(o, m):
                return np.sign(np.cos(o)) * np.abs(np.cos(o))**m
            u = np.linspace(-np.pi, np.pi, N, endpoint=True)
            v = np.linspace(-np.pi/2.0, np.pi/2.0, N, endpoint=True)
            u = np.tile(u, N)
            v = np.repeat(v, N)
            triangles = []

            x = A * g(v, 2.0 / r) * g(u, 2.0 / r)
            y = B * g(v, 2.0 / s) * f(u, 2.0 / s)
            z = C * f(v, 2.0 / t)
            vertices =  np.concatenate([np.expand_dims(x, 1),
                                        np.expand_dims(y, 1),
                                        np.expand_dims(z, 1)], axis=1)
            vertices = vertices @ rotation # apply rotation 

            triangles = []
            for i in range(N-1):
                for j in range(N-1):
                    triangles.append([i*N+j, (i+1)*N+j+1, (i+1)*N+j])
                    triangles.append([i*N+j, i*N+(j+1), (i+1)*N+(j+1)])
            return vertices, triangles

        vertices, triangles = create_superquadric_mesh(scalings[0], scalings[1], scalings[2],
                                                    exponents[0], exponents[1], exponents[2],
                                                    resolution)
        import open3d as o3d
        mesh_sq = o3d.geometry.TriangleMesh()
        mesh_sq.vertices = o3d.utility.Vector3dVector(vertices)
        mesh_sq.triangles = o3d.utility.Vector3iVector(triangles)
        if not os.path.exists("objs"):
            os.makedirs("objs")
        o3d.io.write_triangle_mesh(f"objs/{name}.obj", mesh_sq, write_ascii=True, compressed=False, write_vertex_normals=False, write_vertex_colors=False, write_triangle_uvs=False, print_progress=False)
        
        id_rot_quat = np.array([1,0,0,0])
        scale = np.array([1.0, 1.0, 1.0])
        self.elements[self.__parse_name(name)] = Mesh(f"objs/{name}.obj", translation=translation, rotation=id_rot_quat, scale=scale, color=color, visible=visible)

    def add_arrow(self,
                  name:str,
                  start: np.array,
                  end: np.array,
                  color: np.array=np.array([255, 0, 0]),
                  alpha: float=1.0,
                  stroke_width: float=0.01,
                  head_width: float=0.03,
                  visible: bool=True):
        """Add an arrow."""

        self.elements[self.__parse_name(name)] = Arrow(start, end, color, alpha, stroke_width, head_width, visible)

    def add_motion(self, 
                   name: str, 
                   motion_type: str, 
                   motion_direction: np.array, 
                   motion_origin_pos: np.array,
                   motion_viz_orient: str, 
                   motion_dir_color: np.array=np.array([0, 255, 0]), 
                   motion_origin_color: np.array=np.array([0, 255, 0]), 
                   visible: bool=True):
        """
        Adds a motion vector to the visualizer.

        :param name: 
            Name of the motion vector, which will be displayed in the visualizer.
        :param motion_type: 
            Type of motion: 
            - "trans" for translational motion
            - "rot" for rotational motion
        :param motion_direction: 
            A 3D vector (shape: 3x1, dtype: float32) representing the direction of the motion vector.
        :param motion_origin_pos: 
            A 3D point (shape: 3x1, dtype: float32) representing the origin position of the motion vector.
        :param motion_viz_orient: 
            Orientation of the motion vector visualization. 
            - "outwards": The motion vector points away from the origin.
            - "inwards": The motion vector points towards the origin.
        :param motion_dir_color: 
            RGB color (shape: 3x1, dtype: int32) for the motion vector. 
            Defaults to green ([0, 255, 0]).
        :param motion_origin_color: 
            RGB color (shape: 3x1, dtype: int32) for the origin of the motion vector. 
            Defaults to green ([0, 255, 0]).
        :param visible: 
            Boolean indicating whether the motion vector should be visible in the visualizer. 
            Defaults to True.
        
        :raises AssertionError: 
            If `motion_type` is not one of ["trans", "rot"] or 
            if `motion_viz_orient` is not one of ["outwards", "inwards"].
        """
        assert motion_type in ["trans", "rot"], f"Unknown motion_type option {motion_type}"
        assert motion_viz_orient in ["outwards", "inwards"], f"Unknown motion_viz_orient option {motion_viz_orient}"

        self.elements[self.__parse_name(name)] = Motion(motion_type, motion_origin_pos, motion_direction, motion_viz_orient, motion_dir_color, motion_origin_color, visible)



import bpy
import math
import mathutils
import numpy as np
import subprocess
import json
import os

C = bpy.context
D = bpy.data

import sys
argv = sys.argv
try:
    argv = argv[argv.index("--") + 1:]  # get all args after "--"
except ValueError:
   argv = []

def clear_scene():
  # Remove all meshes from the scene, keep the light and camera
  for o in C.scene.objects:
    print(o)
    if o.type == 'MESH':
        o.select_set(True)
    else:
        o.select_set(False)
  bpy.ops.object.delete()


def render(output_prefix, file_format='PNG', color_mode='RGBA', animation=False):
  """
  :path: the file path of the rendered image
  :file_format: {PNG, JPEG}
  """
  C.scene.render.image_settings.file_format=file_format
  C.scene.render.filepath = output_prefix
  print('output_prefix:', output_prefix)
  
  # C.scene.view_settings.view_transform = 'Standard'
  # D.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (1, 1, 1, 1)
  # bpy.context.scene.world.use_nodes = False
  # bpy.context.scene.world.color = (1, 1, 1)
  # C.scene.render.alpha_mode = 'SKY'
  
  if animation:
    bpy.ops.curve.primitive_bezier_circle_add(radius=1, enter_editmode=False, align='WORLD', location=(0, 0, 0), scale=(1, 1, 1))
    bezier_circle = C.object
    animation_length = 60
    bpy.context.object.data.path_duration = animation_length
    bpy.context.scene.frame_end = animation_length
    bpy.context.scene.render.resolution_y = 406
    bpy.context.scene.render.resolution_x = 720
    bpy.context.scene.cycles.samples = 10
    bpy.data.scenes['Scene'].render.ffmpeg.codec = 'H264'

    light = C.scene.objects['Point']
    light.location = [0.0, 0.0, 0.0]
    light.data.energy = 60
    light.data.shadow_soft_size = 0.1

    cam = C.scene.objects['Camera']
    cam.select_set(True)
    bpy.context.view_layer.objects.active = cam
    cam.matrix_world = mathutils.Matrix(np.eye(4))
    cam.location = [0.0, 0.0, 0.5]
    bpy.ops.object.constraint_add(type='FOLLOW_PATH')
    cam.constraints["Follow Path"].target = bezier_circle
    bpy.ops.object.constraint_add(type='TRACK_TO')
    cam.constraints["Track To"].target = light
    bpy.ops.constraint.followpath_path_animate(constraint="Follow Path", owner='OBJECT')

  bpy.ops.render.render(use_viewport=True, animation=animation, write_still=True)

  if animation:
    output_filepath = output_prefix + '.mp4'
    subprocess.run(["ffmpeg", "-y", "-i", f'{output_prefix}%04d.png', "-vcodec", "libx264", "-vf", "format=yuv420p", "-y", output_filepath])
    subprocess.run(["ffmpeg", "-y", "-i", output_filepath, "-pix_fmt", "rgb24", output_filepath[:-3]+'gif'])


def save_blender_scene(path: str) -> None:
  bpy.ops.wm.save_as_mainfile(filepath=path)


def compute_object_center(object):
  local_bbox_center = 0.125 * sum((mathutils.Vector(b) for b in object.bound_box),
                                   mathutils.Vector())
  return object.matrix_world @ local_bbox_center


def look_at(camera,
            eye=mathutils.Vector((13.0, 13.0, 13.0)),
            at=mathutils.Vector((0.0, 0.0, 3.0)),
            up=mathutils.Vector((0.0, 0.0, 1.0))):
  d = (eye - at).normalized()
  r = up.cross(d).normalized()
  u = d.cross(r).normalized()
  camera.matrix_world = mathutils.Matrix(((r.x, u.x, d.x, eye.x),
                                          (r.y, u.y, d.y, eye.y),
                                          (r.z, u.z, d.z, eye.z),
                                          (0.0, 0.0, 0.0, 1.0)))
  
# def create_video(input_dir, pattern, output_filepath):
#   trans_to_white = "format=yuva444p,\
#   geq=\
#   'if(lte(alpha(X,Y),16),255,p(X,Y))':\
#   'if(lte(alpha(X,Y),16),128,p(X,Y))':\
#   'if(lte(alpha(X,Y),16),128,p(X,Y))'"

#   # f = "color=white,format=rgb24[c];[c][0]scale2ref[c][i];[c][i]overlay=format=auto:shortest=1,setsar=1"
#   # cmd = ["ffmpeg", "-i", pattern, '-filter_complex', f, '-y', output_filepath]
#   # subprocess.run(cmd)

#   import glob
#   for fi in glob.glob(f'{input_dir}/output_*.png'):
#     subprocess.run(["convert", "-flatten", fi, fi])
#     # subprocess.run(["convert", fi, "-background", "white", "-alpha", "remove", "-flatten", "-alpha", "off", fi])
#   # subprocess.run(["ffmpeg", "-i", pattern, '-y', output_filepath])
#   subprocess.run(["ffmpeg", "-i", f'{input_dir}/{pattern}', "-vcodec", "libx264", "-vf", "format=yuv420p", "-y", output_filepath])
#   # subprocess.run(["convert", "-delay", "1", "-loop", "0", "*.png", "myimage.gif"])


def init_scene(resolution_x=800, resolution_y=600):
  # render stuff
  C.scene.render.resolution_x = resolution_x
  C.scene.render.resolution_y = resolution_y
  D.scenes["Scene"].render.film_transparent = True
  C.scene.render.image_settings.color_mode = 'RGBA'
  C.scene.view_settings.look = 'AgX - Medium High Contrast'
  C.scene.view_settings.view_transform = 'Standard'
  C.scene.render.engine = 'CYCLES'
  C.scene.cycles.device = 'GPU'
  C.scene.cycles.preview_samples = 50
  C.scene.cycles.samples = 50

  # Add lights
  C.scene.objects['Light'].data.shadow_soft_size = 1.0
  C.scene.objects['Light'].data.cycles.cast_shadow = False
  bpy.ops.object.light_add(type='POINT', radius=1, align='WORLD', location=(-1, 1, 10), scale=(1, 1, 1))
  C.scene.objects['Point'].data.energy = 7000.0
  C.scene.objects['Point'].data.shadow_soft_size = 3

  for i in range(4):
    bpy.ops.object.light_add(type='POINT', radius=0.1, align='WORLD', location=(i%2, i//2, 1), scale=(1, 1, 1))
    C.scene.objects[f'Point.{str(i+1).zfill(3)}'].data.energy = 5.0
    C.scene.objects[f'Point.{str(i+1).zfill(3)}'].data.shadow_soft_size = 0.1
    C.scene.objects[f'Point.{str(i+1).zfill(3)}'].data.color = (1, 0.795182, 0.375358)


def create_mat(obj, color=None):
    mat = bpy.data.materials.new(name="test")
    mat.use_backface_culling = True
    obj.data.materials.append(mat)
    mat.use_nodes = True
    mat.node_tree.nodes.new(type="ShaderNodeVertexColor")
    mat.node_tree.nodes["Principled BSDF"].inputs[7].default_value = 0  # specular
    mat.node_tree.nodes["Principled BSDF"].inputs[12].default_value = 0  #
    if color:
      print('mesh color', color)
      mat.node_tree.nodes["Principled BSDF"].inputs[0].default_value = (color[0]/255.0, color[1]/255.0, color[2]/255.0, 1.0)
    else:
      mat.node_tree.nodes["Color Attribute"].layer_name = "Col"
      mat.node_tree.links.new(
        mat.node_tree.nodes["Principled BSDF"].inputs["Base Color"],
        mat.node_tree.nodes["Color Attribute"].outputs["Color"])


def cylinder_between(x1, y1, z1, x2, y2, z2, r, color, alpha):
    dx = x2 - x1
    dy = y2 - y1
    dz = z2 - z1    
    dist = math.sqrt(dx**2 + dy**2 + dz**2)
    bpy.ops.mesh.primitive_cylinder_add(
        radius = r, 
        depth = dist,
        location = (dx/2 + x1, dy/2 + y1, dz/2 + z1)   
    ) 
    phi = math.atan2(dy, dx) 
    theta = math.acos(dz/dist) 
    C.object.rotation_euler[1] = theta 
    C.object.rotation_euler[2] = phi

    mat = bpy.data.materials.new(name="test")
    mat.use_backface_culling = True
    C.object.data.materials.append(mat)
    mat.use_nodes = True
    mat.node_tree.nodes["Principled BSDF"].inputs[0].default_value = (color[0]/255.0, color[1]/255.0, color[2]/255.0, alpha)
    return C.object


def main():
    clear_scene()
    init_scene()

    path_json = f'nodes.json'
    with open(path_json) as f:
        nodes_dict = json.load(f)
    animation = False
    for name, properties in nodes_dict.items():
        print(name, properties)
        if properties['type'] == 'points':
           bpy.ops.wm.ply_import(filepath=name+'.ply')
           bpy.ops.object.shade_smooth()
           obj = bpy.data.objects[name]
           create_mat(obj)

        if properties['type'] == 'camera':
           eye = mathutils.Vector(properties['position'])
           at = mathutils.Vector(properties['look_at'])
           up = mathutils.Vector(properties['up'])
           look_at(C.scene.objects['Camera'], eye, at, up)
           C.scene.objects['Camera'].data.lens = properties['focal_length']
           animation = properties['animation']
        
        if properties['type'] == 'cuboid':
           obj = bpy.ops.mesh.primitive_cube_add(size=1, enter_editmode=False, align='WORLD',
                                           location=mathutils.Vector(properties['position']),
                                           scale=mathutils.Vector(properties['size']))
        
        if properties['type'] == 'polyline':
            if len(properties['positions']) <= 1:
                continue
            for i in range(len(properties['positions']) - 1):
                x1 = properties['positions'][i][0]
                y1 = properties['positions'][i][1]
                z1 = properties['positions'][i][2]
                x2 = properties['positions'][i + 1][0]
                y2 = properties['positions'][i + 1][1]
                z2 = properties['positions'][i + 1][2]
                obj = cylinder_between(x1, y1, z1, x2, y2, z2, properties['edge_width'] * 2, properties['color'], properties['alpha'])
        
        if properties['type'] == 'mesh':
          if properties['filename'].split('.')[-1] == 'ply':
            bpy.ops.wm.ply_import(filepath=properties['filename'], forward_axis='Y', up_axis='Z')
          if properties['filename'].split('.')[-1] == 'obj':
            bpy.ops.wm.obj_import(filepath=properties['filename'], forward_axis='Y', up_axis='Z')          
          bpy.ops.object.shade_smooth()
          obj = bpy.data.objects[properties['filename'].split('.')[0]]
          obj.scale = [properties['scale'][0], properties['scale'][1], properties['scale'][2]]
          obj.rotation_mode = 'QUATERNION'  # blender quats are WXYZ
          obj.rotation_quaternion = [properties['rotation'][3], properties['rotation'][0], properties['rotation'][1], properties['rotation'][2]]
          obj.location = [properties['translation'][0], properties['translation'][1], properties['translation'][2]]
          try:
            create_mat(obj, properties['color'])
          except KeyError:
            create_mat(obj)
        # somwhere here parse the blender parameters

    # Render if output filename is provided
    if len(argv) > 0:
        print(argv)
        render(os.path.abspath(argv[0]), animation=animation)

    output_blender_file = f'blender_scene.blend'
    output_blender_file = os.path.abspath(output_blender_file)
    save_blender_scene(output_blender_file)
    print('Saved blender file to:', output_blender_file)



"""Points class i.e. point cloud."""


class Points:
    """Set of points defined by positions, colors, normals and more."""

    def __init__(self, positions, colors, normals, point_size, resolution, visible, alpha, shading_type=1):
        self.positions = positions
        self.colors = colors
        self.normals = normals
        self.point_size = point_size
        self.resolution = resolution
        self.visible = visible
        self.alpha = alpha
        self.shading_type = shading_type

    def get_properties(self, binary_filename):
        """
        :return: A dict conteining object properties. They are written into json and interpreted by javascript.
        """
        json_dict = {
            'type': 'points',
            'visible': self.visible,
            'alpha': self.alpha,
            'shading_type': self.shading_type,
            'point_size': self.point_size,
            'num_points': self.positions.shape[0],
            'binary_filename': binary_filename}
        return json_dict

    def write_binary(self, path):
        """Write points to binary file."""
        bin_positions = self.positions.tobytes()
        bin_normals = self.normals.tobytes()
        bin_colors = self.colors.tobytes()
        with open(path, "wb") as f:
            f.write(bin_positions)
            f.write(bin_normals)
            f.write(bin_colors)

    def write_blender(self, path):
        import open3d as o3d
        import tqdm
        pcd_combined = o3d.geometry.TriangleMesh()

        def _create_sphere_at_xyz(xyz, colors, radius, resolution):
            sphere = o3d.geometry.TriangleMesh.create_sphere(radius=radius, resolution=resolution)
            sphere.compute_vertex_normals()
            sphere.paint_uniform_color(colors)
            sphere = sphere.translate(xyz)
            return sphere

        # Make no-ob points whiter
        # pcd[np.all(pcd[:, 6:] == 120, axis=1), 6:] = [220., 220., 220.]

        for i in range(self.positions.shape[0]):
            pcd_combined += _create_sphere_at_xyz(self.positions[i],
                                                  self.colors[i] / 255.0,
                                                  self.point_size / 1000.0,
                                                  resolution=self.resolution)

        o3d.io.write_triangle_mesh(path, pcd_combined)



"""Camera related functionality."""
import numpy as np


class Camera:
    """Camera class looking at the scene"""

    def __init__(self, position, look_at, up, focal_length, animation):
        self.position = position.astype(np.float32)
        self.look_at = look_at.astype(np.float32)
        self.up = up.astype(np.float32)
        self.focal_length = float(focal_length)
        self.animation = animation

    def get_properties(self, binary_filename):
        """Get the camera properties, they are written into json and interpreted by javascript.

        :return: A dict conteining object properties.
        """
        json_dict = {
            'type': 'camera',
            'position': self.position.tolist(),
            'look_at': self.look_at.tolist(),
            'up': self.up.tolist(),
            'focal_length': self.focal_length,
            'animation': self.animation,
            }
        return json_dict

    def write_binary(self, path):
        """Write camera to binary file."""
        bin_position = self.position.tobytes()
        with open(path, "wb") as f:
            f.write(bin_position)

    def write_blender(self, path):
        # all info is on the json, nothing needs to be done
        pass


"""Lines class e.g. to visualize normals."""
import numpy as np


class Lines:
    """Set of line segments defined by startint points and ending points."""

    def __init__(self, lines_start, lines_end, colors_start, colors_end, visible):
        # Interleave start and end positions for WebGL.
        self.num_lines = lines_start.shape[0]
        self.positions = np.empty((self.num_lines * 2, 3), dtype=lines_start.dtype)
        self.positions[0::2] = lines_start
        self.positions[1::2] = lines_end
        self.colors = np.empty((self.num_lines * 2, 3), dtype=np.uint8)
        self.colors[0::2] = colors_start
        self.colors[1::2] = colors_end
        self.visible = visible

    def get_properties(self, binary_filename):
        """ Get line properties, they are written into json and interpreted by javascript.
        :return: A dict conteining object properties.
        """
        json_dict = {
            'type': 'lines',
            'visible': self.visible,
            'num_lines': self.num_lines,
            'binary_filename': binary_filename}
        return json_dict

    def write_binary(self, path):
        """Write lines to binary file."""

        bin_positions = self.positions.tobytes()
        bin_colors = self.colors.tobytes()
        with open(path, "wb") as f:
            f.write(bin_positions)
            f.write(bin_colors)

    def write_blender(self, path):
        print(type(self).__name__ + '.write_blender() not yet implemented.' )
        return


# Cuboid class e.g. a bounding box.


class Polyline:
    def __init__(self, positions,color, alpha, edge_width, visible):
        self.positions = positions
        self.color = color
        self.alpha = alpha
        self.edge_width = edge_width
        self.visible = visible

    def get_properties(self, binary_filename):
        """ Get line properties, they are written into json and interpreted by javascript.
        :return: A dict conteining object properties.
        """
        json_dict = {
            'type': 'polyline',
            'positions': self.positions.tolist(),
            'color': self.color.tolist(),
            'alpha': float(self.alpha),
            'edge_width': float(self.edge_width),
            'visible': self.visible,
        }
        return json_dict

    def write_binary(self, path):
        """Write lines to binary file."""
        return

    def write_blender(self, path):
        return